#include "../config.h"
#include "test.h"
#include <string.h>
#include <stdio.h>
#include <errno.h>

/* #undef HAVE_STRNUP */
#ifndef HAVE_STRNDUP
char * strndup(const char * s, size_t n) {
   char *d;
   size_t nl;

   if (n == 0) {
     d = malloc(1);
     if (d) {
       *d = 0;
     }
   } else {
     if (!s) {
       return 0;
     }
     nl=strlen(s);
     nl = nl <= n ? nl : n;
     d=malloc(nl+1);
     if (d) {
       d = strncpy(d,s,nl);
       d[nl] = 0;
     }
   }
   return(d);
}
#endif

GString *
au_gstring_to_netstring(GString *s) {
  /* Converts the contents of a GString into a djb-style netstring, why
  ** reinvent the wheel? 
  */
  GString *ns;
  ns = g_string_new("");
  g_string_sprintfa(ns,"%lu:",s->len);
  ns = g_string_append(ns,s->str);
  ns = g_string_append(ns,",");
  return ns;
}

GString *
au_netstring_to_gstring(GString *ns) {
  GString *str;
  size_t ns_len;
  char *tnbuf;
  char *tbuf;
  str = g_string_new("");
  sscanf(ns->str,"%d:",&ns_len);
  tnbuf = strchr(ns->str,(int)':')+1;
  tbuf = (char*)strndup(tnbuf,ns_len); /* strndup is a GNU extention, this is
				   just for prototyping. */
  str = g_string_append(str,tbuf);
  free(tbuf);
  return str;
}

GString *
au_gboolean_to_netstring(gboolean b) {
  GString *ns;
  ns = g_string_new("");
  g_string_sprintfa(ns,"%lu:",1);
  ns = g_string_append(ns,b==TRUE?"1":"0");
  ns = g_string_append(ns,",");
  return ns;
}

gboolean
au_netstring_to_gboolean(GString *ns) {
  GString *str;
  size_t ns_len;
  char *tnbuf;
  gint tval;
  gboolean result;
  str = g_string_new("");
  sscanf(ns->str,"%d:",&ns_len);
  tnbuf = strchr(ns->str,(int)':')+1;
  sscanf(tnbuf,"%d",&tval);
  result = tval == 0 ? FALSE : TRUE;
  return result;
}

gint
au_netstring_to_gint(GString *ns) {
  GString *str;
  size_t ns_len;
  char *tnbuf;
  gint tval;
  gint result;
  str = g_string_new("");
  sscanf(ns->str,"%d:",&ns_len);
  tnbuf = strchr(ns->str,(int)':')+1;
  sscanf(tnbuf,"%d",&tval);
  result = tval;
  g_string_free(str,TRUE);
  return result;
}

gulong
au_netstring_to_gulong(GString *ns) {
  GString *str;
  size_t ns_len;
  char *tnbuf;
  gulong tval;
  gulong result;
  str = g_string_new("");
  sscanf(ns->str,"%d:",&ns_len);
  tnbuf = strchr(ns->str,(int)':')+1;
  sscanf(tnbuf,"%d",&tval);
  result = tval;
  g_string_free(str,TRUE);
  return result;
}

gdouble
au_netstring_to_gdouble(GString *ns) {
  GString *str,*tmpstr;
  size_t ns_len;
  char *tnbuf;
  gdouble tval;
  gdouble result;
  str = g_string_new("");
  sscanf(ns->str,"%d:",&ns_len);
  tmpstr = g_string_new(strchr(ns->str,(int)':')+1);
  g_string_truncate(tmpstr,tmpstr->len-1);
  sscanf(tmpstr->str,"%lf",&tval);
  result = tval;
  g_string_free(str,TRUE);
  g_string_free(tmpstr,TRUE);
  return result;
}

GString *
au_gint_to_netstring(gint gi) {
  GString *gint_str;
  GString *ns;
  gint_str = g_string_new("");
  g_string_sprintf(gint_str,"%d",gi);
  ns = g_string_new("");
  g_string_sprintfa(ns,"%lu:",gint_str->len);
  ns = g_string_append(ns,gint_str->str);
  ns = g_string_append(ns,",");
  g_string_free(gint_str,TRUE);
  return ns;
}


GString *
au_gdouble_to_netstring(gdouble gd) {
  GString *gdouble_str;
  GString *ns;
  gdouble_str = g_string_new("");
  g_string_sprintf(gdouble_str,"%f",gd);
  ns = g_string_new("");
  g_string_sprintfa(ns,"%lu:",gdouble_str->len);
  ns = g_string_append(ns,gdouble_str->str);
  ns = g_string_append(ns,",");
  g_string_free(gdouble_str,TRUE);
  return ns;
}

GString *
au_gulong_to_netstring(gulong gul) {
  GString *gul_str;
  GString *ns;
  gul_str = g_string_new("");
  g_string_sprintf(gul_str,"%lu",gul);
  ns = g_string_new("");
  g_string_sprintfa(ns,"%lu:",gul_str->len);
  ns = g_string_append(ns,gul_str->str);
  ns = g_string_append(ns,",");
  g_string_free(gul_str,TRUE);
  return ns;
}


GString *
au_read_netstring(int fd) {
  /* Read a netstring from a file descriptor into a GString
     returns the contents of the netstring, not the netstring "wrapper".

     djb's sample code: (assumes you're reading from stdin, is here for
     reference)

      if (scanf("%9lu",&len) < 1) barf();  /* >999999999 bytes is bad * /
      if (getchar() != ':') barf();
      buf = malloc(len + 1);       /* malloc(0) is not portable * /
      if (!buf) barf();
      if (fread(buf,1,len,stdin) < len) barf();
      if (getchar() != ',') barf();
   */
  FILE *pipestream;
  GString *tns;
  gulong len;
  pipestream = fdopen(fd,"r");
  if (!pipestream) {
    fprintf(stderr,"fatal: %s\n",strerror(errno));
    abort();
  }
  if (fscanf(pipestream,"%9lu",&len) < 1) {
    g_string_free(tns,TRUE);
    return 0;
  }
  tns = g_string_sized_new(len+1);
  if (':' != fgetc(pipestream)) {
    g_string_free(tns,TRUE);
    return 0;
  }
  if (fread(tns->str,1,len,pipestream) < len) {
    /* this is really not friendly in the case that the length is 
       quite long and goes over 4K, I'm just going to punt on that
       because I don't expect many test status results to go over that. */
    g_string_free(tns,TRUE);
    return 0;
  }
  if (',' != fgetc(pipestream)) {
    g_string_free(tns,TRUE);
    return 0;
  }
  fclose(pipestream);
  tns->str[len] = 0;
  tns->len = len;
  return tns;
}

GString *
au_pop_netstring(GString *ns) {
  /* Returns the netstring at the front of the GString, and changes
   * the input GString.  This API may change.  I'm not proud of this code.
   */
  GString *tns;
  gulong str_len;
  gint digit_len;
  char *digits;
  char *popped;
  GString *token_tmp;  
  int copy_len;

  token_tmp = g_string_new(ns->str);

  digits = strtok(token_tmp->str,":");

  if (!digits) {
    g_string_free(token_tmp,TRUE);
    return 0;
  }
  digit_len = strlen(digits);
  str_len = strtoul(digits,0,10);
  if (str_len > 9999999) {
    g_string_free(token_tmp,TRUE);
    return 0;
  }

  /* length ':' string ',' \0 */
  tns = g_string_sized_new(digit_len+1+str_len+1+1);
  popped = (char*)strndup(ns->str,digit_len+1+str_len+1);
  g_string_append(tns,popped);
  g_string_erase(ns,0,digit_len+1+str_len+1);
  free(popped);
  return tns;
}
