/* $Id: c-unit-suite.c,v 1.9 2001/05/06 02:20:08 spj Exp $
** ----------------------------------------------------------------------
** $Log: c-unit-suite.c,v $
** Revision 1.9  2001/05/06 02:20:08  spj
** A nice bugfix and some more documentation.
**
** Revision 1.8  2001/04/29 22:39:13  spj
** *** empty log message ***
**
** Revision 1.7  2001/03/30 06:31:08  spj
** Memory leak cleanups mostly and a revision bump.  0.08 is NOT backward compatible
** with 0.07 because of a change in au_new_test().
** --spj
**
** Revision 1.6  2001/03/30 05:15:07  spj
** Serialize and unserialize now work.  Now I want to figure out how I can measure
** the amount of ram being used by the application-- it's pretty obvious to me
** now that during stress testing I've got a huge memory leak which is slowing
** things down... but more importantly... I've got a huge memory leak.
**
** Question is, do I write the fork() code before I fix the leak? If I write the fork() code
** it will likely hide the leakage. --spj
**
** Revision 1.5  2001/03/25 19:44:35  spj
** Rather ugly changes, but at the moment au_pop_netstring passes it's unit
** tests, so I'm closer to finishing au_test_unserialize.
** --spj
**
** Revision 1.4  2001/03/10 19:23:09  spj
** Split off netstrings from the test.c module into au-netsring.c.
**
** Removing the bit of hand-coded test suite setup and making it initialize
** from the same static array of test cases caused the stress-test to turn
** up a bug in au_new_test where it wasn't initizlizing the elapsed times
** for the test. I guess I omitted or ignored that test in the stress
** rounds.  Now stress tests take a little bit of time which means that
** I should have paid more attention to the errors. :)
**
** Revision 1.3  2001/03/10 06:03:36  spj
** Reduced amount of "hand coded" test setup in c-unit-suite because
** it was becoming onerous to repeat the code. W/e you find yourself
** copying and pasting code, think about perhaps making your life easier.
** --spj
**
** Revision 1.2  2001/03/08 06:35:50  spj
** Tested netstring functions.  Serialize is tested, unserialize needs to be
** written.
**
** Revision 1.1  2001/03/02 04:51:07  spj
** c-unit-suite moved from test to c-unit dir.; c-unit texinfo documentation renamed / re-outlined for future language support
**
** Revision 1.3  2001/02/23 05:10:22  spj
** Started a Texinfo manual document, and added the beginning of setup/teardown
** functions which are now invoked around the test.
**
** Revision 1.2  2001/02/19 19:26:17  spj
** 0.04 checkin.  Some report reformats and new unit tests for au_remove_test
** and au_delete_test.  TODO: au_delete_testcase.
** --spj
**
** Revision 1.1.1.1  2001/02/18 20:42:40  spj
** First checkin.  'c-unit' works but self-testing suite needs to be extended
** to cover more of its functionality. "self-testing" recursively with your
** own self-testing framework gets confusing sometimes. ;) --spj
**
**
*/

#include <c-unit/test.h>
#include <glib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>

autounit_testcase_t *b_tc;
autounit_test_t     *b_t;

gint 
test_sample_test(autounit_test_t *t) {
}

gint
cus_setup_testcase(autounit_test_t *t) {
  b_tc = au_new_testcase(g_string_new("Mindbending Self Test"),0,0);
  b_t = au_new_test(g_string_new("test_sample_test"),test_sample_test);
}

gint
cus_teardown_testcase(autounit_test_t *t) {
  au_delete_test(b_t);
  au_delete_testcase(b_tc);
}

gint
test_au_new_testcase(autounit_test_t *t) {
  autounit_testcase_t *x_tc = 
    au_new_testcase(g_string_new("Mindbending Self Test"),
		    0,0);
  au_assert(t,"testcase_name set incorrectly",
	    strcmp(x_tc->testcase_name->str,"Mindbending Self Test")==0); 
  au_assert(t,"testcase_status not blank on virgin struct",
	    strcmp(x_tc->testcase_status->str,"")==0);
  au_assert(t,"testcase_pct_complete not 0% on virgin struct",
	    x_tc->testcase_pct_complete==0);
  au_assert(t,"testcase_seconds_elapsed not 0 on virgin struct",
	    x_tc->testcase_seconds_elapsed==0);
  au_assert(t,"testcase_useconds_elapsed not 0 on virgin struct",
	    x_tc->testcase_useconds_elapsed==0);
  au_assert(t,"tests not NULL on virgin struct",x_tc->tests == NULL); 
  au_delete_testcase(x_tc);
}


gint
test_au_new_test(autounit_test_t *t) {
  autounit_test_t *b_t2;
  /* delete memory leak before running this test. */
  au_remove_test(b_tc,b_t);
  au_delete_test(b_t);
  au_assert(t,"testcase should have empty list",
	    g_slist_length(b_tc->tests)==0);
  b_t = au_new_test(g_string_new("test_sample_test"),
		test_sample_test);
  b_tc = au_add_test(b_tc,b_t);

  au_assert(t,"different test list length than expected",
	    g_slist_length(b_tc->tests)==1);
  b_t2 = au_new_test(g_string_new("test_sample_test"),
		test_sample_test);
  b_tc = au_add_test(b_tc,b_t2);
  au_assert(t,"different test list length than expected",
	    g_slist_length(b_tc->tests)==2);
  au_assert(t,"test_name set incorrectly",
	    strcmp(b_t->test_name->str,"test_sample_test")==0); 
  au_assert(t,"test_run not FALSE on virgin struct",
	    b_t->test_run==FALSE);
  au_assert(t,"failed_assertions not 0 on virgin struct",
	    b_t->failed_assertions==0);
  au_assert(t,"test_seconds_elapsed not 0 on virgin struct",
	    b_t->test_seconds_elapsed==0.0);
  au_assert(t,"test_useconds_elapsed not 0 on virgin struct",
	    b_t->test_useconds_elapsed==0);
  au_assert(t,"testcase parent not the one we expected",
	    b_t->testcase == b_tc);
  au_assert(t,"default do_fork set incorrectly",
	    b_t->do_fork == TRUE); 
  /* clean up */
  au_remove_test(b_tc,b_t);
  au_remove_test(b_tc,b_t2);
  /* teardown will delete b_tc and b_t, but not b_t2 */
  au_delete_test(b_t2);
}

gint
test_au_add_test(autounit_test_t *t) {
  au_add_test(b_tc,b_t);
  au_assert(t,"different test list length than expected",
	    g_slist_length(b_tc->tests)==1);
  au_add_test(b_tc,b_t);
  au_assert(t,"different test list length than expected",
	    g_slist_length(b_tc->tests)==2);
}

gint
test_au_set_fork_mode(autounit_test_t *t) {
  au_assert(t,"default fork mode",
	    b_t->do_fork == TRUE);
  au_set_fork_mode(b_t,FALSE);
  au_assert(t,"changed fork mode (FALSE)",
	    b_t->do_fork == FALSE);
  au_set_fork_mode(b_t,TRUE);
  au_assert(t,"changed fork mode (TRUE)",
	    b_t->do_fork == TRUE);
}

gint
test_au_remove_test(autounit_test_t *t) {
  au_add_test(b_tc,b_t);
  au_assert(t,"length of tests list is not 1 after add",
	    g_slist_length(b_tc->tests)==1);
  au_remove_test(b_tc,b_t);
  au_assert(t,"length of tests list is not 1 after one deletion",
	    g_slist_length(b_tc->tests)==0);
  au_remove_test(b_tc,b_t);
}

gint
test_au_gstring_to_netstring(autounit_test_t *t) {
  GString *ns; 
  ns = au_gstring_to_netstring(g_string_new("1234567890:::abc"));
  au_assert(t,"ns len doesn't match what we expect",
	    ns->len==20);
  au_assert(t,"ns content doesn't match what we expect",
	    strcmp(ns->str, "16:1234567890:::abc,")==0);
  g_string_free(ns,TRUE);
}

gint
test_au_netstring_to_gstring(autounit_test_t *t) {
  GString *ns; 
  GString *str; 
  ns = au_gstring_to_netstring(g_string_new("1234567890:::abc"));
  str = au_netstring_to_gstring(ns);
  au_assert(t,"str len doesn't match what we expect",
	    str->len==16);
  au_assert(t,"str content doesn't match what we expect",
	    strcmp(str->str, "1234567890:::abc")==0);
  g_string_free(ns,TRUE);
  g_string_free(str,TRUE);
}

gint
test_au_gboolean_to_netstring(autounit_test_t *t) {
  GString *ns; 
  gboolean b_TRUE = TRUE;
  gboolean b_FALSE = FALSE;

  ns = au_gboolean_to_netstring(b_TRUE);
  au_assert(t,"ns len doesn't match what we expect",
	    ns->len==4);
  au_assert(t,"ns content doesn't match what we expect",
	    strcmp(ns->str, "1:1,")==0);
  g_string_free(ns,TRUE);

  ns = au_gboolean_to_netstring(b_FALSE);
  au_assert(t,"ns len doesn't match what we expect",
	    ns->len==4);
  au_assert(t,"ns content doesn't match what we expect",
	    strcmp(ns->str, "1:0,")==0);
  g_string_free(ns,TRUE);
}

gint
test_au_netstring_to_gboolean(autounit_test_t *t) {
  gboolean b_TRUE = TRUE;
  gboolean b_FALSE = FALSE;
  GString *ns_FALSE; 
  GString *ns_TRUE; 
  gboolean v_TRUE;
  gboolean v_FALSE;

  ns_TRUE = au_gboolean_to_netstring(b_TRUE);
  ns_FALSE = au_gboolean_to_netstring(b_FALSE);
  v_TRUE = au_netstring_to_gboolean(ns_TRUE);
  au_assert(t,"gboolean not TRUE",v_TRUE == TRUE);
  v_FALSE = au_netstring_to_gboolean(ns_FALSE);
  au_assert(t,"gboolean not FALSE",v_FALSE == FALSE);

  g_string_free(ns_FALSE,TRUE);
  g_string_free(ns_TRUE,TRUE);
}

gint
test_au_netstring_to_gint(autounit_test_t *t) {
  gint fourtytwo = 42;
  GString *ns_fourtytwo; 
  gint v_42;

  ns_fourtytwo = au_gint_to_netstring(fourtytwo);
  v_42 = au_netstring_to_gint(ns_fourtytwo);
  au_assert(t,"gint not 42",v_42 == 42);

  g_string_free(ns_fourtytwo,TRUE);
}

gint
test_au_netstring_to_gulong(autounit_test_t *t) {
  gint fourtytwoM = 42000000;
  GString *ns_fourtytwoM; 
  gint v_42M;

  ns_fourtytwoM = au_gint_to_netstring(fourtytwoM);
  v_42M = au_netstring_to_gulong(ns_fourtytwoM);
  au_assert(t,"gulong not 42000000",v_42M == 42000000);

  g_string_free(ns_fourtytwoM,TRUE);
}

gint
test_au_netstring_to_gdouble(autounit_test_t *t) {
  gdouble nnone = 99.1;
  GString *ns_nnone; 
  gdouble v_nnone;

  ns_nnone = au_gdouble_to_netstring(nnone);
  v_nnone = au_netstring_to_gdouble(ns_nnone);
  au_assert(t,"double not 99.1",v_nnone == 99.1);
  
  g_string_free(ns_nnone,TRUE);
}


gint
test_au_gint_to_netstring(autounit_test_t *t) {
  GString *ns; 
  gint b_42 = 42;
  gint b_n42 = -42;

  ns = au_gint_to_netstring(b_42);
  au_assert(t,"ns len doesn't match what we expect",
	    ns->len==5);
  au_assert(t,"ns content doesn't match what we expect",
	    strcmp(ns->str, "2:42,")==0);
  g_string_free(ns,TRUE);

  ns = au_gint_to_netstring(b_n42);
  au_assert(t,"ns len doesn't match what we expect",
	    ns->len==6);
  au_assert(t,"ns content doesn't match what we expect",
	    strcmp(ns->str, "3:-42,")==0);
  g_string_free(ns,TRUE);
}

gint
test_au_gdouble_to_netstring(autounit_test_t *t) {
  GString *ns; 
  gdouble b_pi = 4.5; /* Ok, it's a really bad approximation. */
  gdouble b_npi = -b_pi; /* Some bad arithmetic or lossage happens
			 **  here so I just fudged around it in the
			 **  "expected" results.  The point here is
			 **  testing whether the netstring is what
			 **  we're expecting, not the arithmetic.  I
			 **  just know this is going to blow-up
			 **  sometime when someone ports it somewhere
			 **  else.  It'll give someone some practise
			 **  using unit tests. ;) */

  ns = au_gdouble_to_netstring(b_pi);
  au_assert(t,"ns len doesn't match what we expect",
	    ns->len==11);
  au_assert(t,"ns content doesn't match what we expect",
	    strcmp(ns->str, "8:4.500000,")==0);
  g_string_free(ns,TRUE);

  ns = au_gint_to_netstring(b_npi);
  au_assert(t,"ns len doesn't match what we expect",
	    ns->len==5);
  au_assert(t,"ns content doesn't match what we expect",
	    strcmp(ns->str, "2:-4,")==0);
  g_string_free(ns,TRUE);
}

gint
test_au_gulong_to_netstring(autounit_test_t *t) {
  GString *ns; 
  gulong b_12345678 = 12345678;

  ns = au_gulong_to_netstring(b_12345678);
  au_assert(t,"ns len doesn't match what we expect",
	    ns->len==11);
  au_assert(t,"ns content doesn't match what we expect",
	    strcmp(ns->str, "8:12345678,")==0);
  g_string_free(ns,TRUE);

}

gint
test_au_read_netstring(autounit_test_t *t) {
  GString *hello_ns;
  GString *hello;
  FILE *test_fp;
  hello_ns = au_gstring_to_netstring(g_string_new("**"));
  /*
    printf("\nhello-nsstr = '%s'\n",hello_ns->str); 
    printf("\nhello-nsstr-len = %d\n",hello_ns->len); 
    */

  test_fp = fopen("tst.au_read_netstring","w");
  if (fwrite(hello_ns->str,hello_ns->len,sizeof(char),test_fp) < 0) {
    au_assert(t,"test failed to create data file",0);
    return;
  }
  fclose(test_fp);
  test_fp = fopen("tst.au_read_netstring","r");
  hello = au_read_netstring(fileno(test_fp));
  fclose(test_fp);
  unlink("tst.au_read_netstring");

  /* 
     printf("\nhello-str = '%s'\n",hello->str); 
     printf("\nhello-str-len = %d\n",hello->len); 
     */

  au_assert(t,"hello content doesn't match what we expect",
	    strcmp(hello->str, "**")==0);
  g_string_free(hello_ns,TRUE);
  g_string_free(hello,TRUE);
}

gint
test_au_test_serialize_unserialize(autounit_test_t *t) {
  /* only needs to handle the "results" data, so we don't care about
  ** the testcase pointer and test_fp pointer.
  **/
  GString *str_test;
  GString *read_test;
  autounit_test_t *us_test;
  str_test = au_test_serialize(b_t);
  au_assert(t,"serialized test is what we expect",
	    strcmp(str_test->str,
		   "46:16:test_sample_test,1:0,0:,1:0,8:0.000000,1:0,,")==0);
  /* When the serialized test is read from the child's pipe with au_read_netstream
   * it will have unwrapped the outer netstring encoding already, we'll simulate
   * this here with a call to au_netstring_to_gstring(). */
  read_test = au_netstring_to_gstring(str_test);
  us_test = au_test_unserialize(read_test);
  au_assert(t,"us_test isn't allocated",us_test!=0);
  au_assert(t,"test_name isn't allocated",us_test->test_name!=0);
  au_assert(t,"test_name mismatch",
	    strcmp(b_t->test_name->str,us_test->test_name->str)==0);
  au_assert(t,"test_run mismatch",b_t->test_run == us_test->test_run);
  au_assert(t,"test_status isn't allocated",us_test->test_status!=0);
  au_assert(t,"test_status mismatch",strcmp(b_t->test_status->str,us_test->test_status->str)==0);
  au_assert(t,"failed_assertions mismatch",b_t->failed_assertions == us_test->failed_assertions);
  au_assert(t,"test_seconds_elapsed mismatch",b_t->test_seconds_elapsed == us_test->test_seconds_elapsed);
  au_assert(t,"test_seconds_elapsed mismatch",b_t->test_useconds_elapsed == us_test->test_useconds_elapsed);
  au_delete_test(us_test); 
  g_string_free(read_test,TRUE);
  g_string_free(str_test,TRUE);
}

gint
test_au_pop_netstring(autounit_test_t *t) {
  GString *test_ns;
  GString *test_str;
  GString *test_str2;
  GString *test_str_ns;
  GString *test_str2_ns;
  GString *test_pop;
  
  test_ns = g_string_new("");
  test_str = g_string_new("funk");
  test_str2 = g_string_new("groove");
  test_str_ns = au_gstring_to_netstring(test_str);
  test_str2_ns = au_gstring_to_netstring(test_str2);
  g_string_append(test_ns,test_str_ns->str);
  g_string_append(test_ns,test_str2_ns->str);

  test_pop = au_pop_netstring(test_ns);
  au_assert(t,"test_pop is the first netstring in the 'stack' of netstrings",
	    strcmp(test_pop->str,test_str_ns->str)==0);
  au_assert(t,"test_ns is the remainder",
	    strcmp(test_ns->str,test_str2_ns->str)==0);
  g_string_free(test_pop,TRUE);
  test_pop = au_pop_netstring(test_ns);
  au_assert(t,"test_pop is the last of the string",
	    strcmp(test_pop->str,test_str2_ns->str)==0);
  au_assert(t,"test_pop should be empty now",
	    test_ns->len == 0);
  g_string_free(test_ns,TRUE);
  g_string_free(test_str,TRUE);
  g_string_free(test_str2,TRUE);
  g_string_free(test_str_ns,TRUE);
  g_string_free(test_str2_ns,TRUE);
  g_string_free(test_pop,TRUE);
}

gint
test_au_delete_test(autounit_test_t *t) {
  /* Test may be  obselete now that setup/teardown call this function?
   * or do I need a flag from the test to notify the test case that it should
   * not call the teardown and/or setup function?
   */

  if (0) {
    au_remove_test(b_tc,b_t);
    au_delete_test(b_t);
  }
}

#ifndef HAVE_STRNDUP
gint
test_builtin_strndup(autounit_test_t *t) {
  char *original = "asdf1234";
  char *short_cpy;
  char *long_cpy;
  char *empty_cpy;

  short_cpy = strndup(original,(size_t)4);
  au_assert(t,"short_cpy",
	    strcmp(short_cpy,"asdf") == 0);
  
  long_cpy = strndup(original,(size_t)16);
  au_assert(t,"long_cpy",
	    strcmp(long_cpy,"asdf1234") == 0);
  empty_cpy = strndup(original,(size_t)0);
  au_assert(t,"empty_cpy",
	    strcmp(empty_cpy,"") == 0);
}
#endif // HAVE_STRNDUP

typedef struct {
  char *name;
  autounit_test_fp_t test;
  gboolean enabled;
  gboolean forking;
}  test_link_t;

static test_link_t tests[] = {
#ifndef HAVE_STRNDUP
  {"au_builtin_strndup", test_builtin_strndup, TRUE, TRUE},
#endif
  {"au_new_testcase", test_au_new_testcase, TRUE, TRUE},
  {"au_new_test", test_au_new_test, TRUE, TRUE},
  {"au_set_fork_mode", test_au_set_fork_mode, TRUE, FALSE},
  {"au_add_test", test_au_add_test, TRUE, TRUE},
  {"au_remove_test", test_au_remove_test, TRUE, TRUE},
  {"au_test_gstring_to_netstring", test_au_gstring_to_netstring, TRUE, TRUE},
  {"au_test_netstring_to_gstring", test_au_netstring_to_gstring, TRUE, TRUE},
  {"au_test_gboolean_to_netstring", test_au_gboolean_to_netstring, TRUE, TRUE},
  {"au_test_gboolean_to_netstring", test_au_gboolean_to_netstring, TRUE, TRUE},
  {"au_test_netstring_to_gboolean", test_au_netstring_to_gboolean, TRUE, TRUE},
  {"au_test_netstring_to_gint", test_au_netstring_to_gint, TRUE, TRUE},
  {"au_test_netstring_to_gulong", test_au_netstring_to_gulong, TRUE, TRUE},
  {"au_test_netstring_to_gdouble", test_au_netstring_to_gdouble, TRUE, TRUE},
  {"au_test_gint_to_netstring",     test_au_gint_to_netstring, TRUE, TRUE},
  {"au_test_gdouble_to_netstring",     test_au_gdouble_to_netstring, TRUE, TRUE},
  {"au_test_gulong_to_netstring",     test_au_gulong_to_netstring, TRUE, TRUE},
  {"au_test_read_netstring",     test_au_read_netstring, TRUE, TRUE},
  {"au_test_au_pop_netstring",     test_au_pop_netstring, TRUE, TRUE},
  {"serialize/deserialize",     test_au_test_serialize_unserialize, TRUE, TRUE},
  {0, 0, FALSE, FALSE}
};

int
main() {
  autounit_testcase_t *c_unit_test_suite;
  autounit_testcase_t *c_unit_stresstest_suite;
  int test_no;
  gint result;
  autounit_test_t *tmp_test;

  c_unit_test_suite = 
    au_new_testcase(g_string_new("Autounit C Unit Self Test"), 
		    cus_setup_testcase,cus_teardown_testcase);

  test_no = 0;
  while (tests[test_no].name != 0) {
    if (tests[test_no].enabled == TRUE) {
      tmp_test = au_new_test(g_string_new(tests[test_no].name),
			     tests[test_no].test);
      au_set_fork_mode(tmp_test,tests[test_no].forking);
      au_add_test(c_unit_test_suite,tmp_test);
    } else {
      /* Warn user/programmer that a test has been disabled. */
      printf("?! '%s' test disabled\n",tests[test_no].name);
    }
    test_no++;
  }

  c_unit_stresstest_suite = 
    au_new_testcase(g_string_new("Autounit C Unit Self Stress Test"),
		    cus_setup_testcase,cus_teardown_testcase);
  test_no = 0;
  while (tests[test_no].name != 0) {
    if (tests[test_no].enabled == TRUE) {
      tmp_test = au_new_test(g_string_new(tests[test_no].name),
			     tests[test_no].test);
      au_set_fork_mode(tmp_test,tests[test_no].forking);
      au_add_test(c_unit_stresstest_suite,tmp_test);
    }
    test_no++;
  }
  result = au_run_testcase(c_unit_test_suite);

  if (!result) {
    result = au_run_stress_testcase(c_unit_stresstest_suite,25,5);
  }
  return result;
}

