/* $Id: test.c,v 1.14 2001/05/06 02:20:08 spj Exp $
** $Log: test.c,v $
** Revision 1.14  2001/05/06 02:20:08  spj
** A nice bugfix and some more documentation.
**
** Revision 1.13  2001/04/30 23:20:27  spj
** Bump to 0.09.
**
** Revision 1.12  2001/04/29 22:39:13  spj
** *** empty log message ***
**
** Revision 1.11  2001/03/30 06:31:08  spj
** Memory leak cleanups mostly and a revision bump.  0.08 is NOT backward compatible
** with 0.07 because of a change in au_new_test().
** --spj
**
** Revision 1.10  2001/03/30 05:15:07  spj
** Serialize and unserialize now work.  Now I want to figure out how I can measure
** the amount of ram being used by the application-- it's pretty obvious to me
** now that during stress testing I've got a huge memory leak which is slowing
** things down... but more importantly... I've got a huge memory leak.
**
** Question is, do I write the fork() code before I fix the leak? If I write the fork() code
** it will likely hide the leakage. --spj
**
** Revision 1.9  2001/03/25 19:44:35  spj
** Rather ugly changes, but at the moment au_pop_netstring passes it's unit
** tests, so I'm closer to finishing au_test_unserialize.
** --spj
**
** Revision 1.8  2001/03/14 03:13:58  spj
** Needed to remove the implicit "au_add_test()" from "au_new_test()" because it's
** not obvious that a "new" operation is going to do that.
**
** Revision 1.7  2001/03/10 19:23:09  spj
** Split off netstrings from the test.c module into au-netsring.c.
**
** Removing the bit of hand-coded test suite setup and making it initialize
** from the same static array of test cases caused the stress-test to turn
** up a bug in au_new_test where it wasn't initizlizing the elapsed times
** for the test. I guess I omitted or ignored that test in the stress
** rounds.  Now stress tests take a little bit of time which means that
** I should have paid more attention to the errors. :)
**
** Revision 1.6  2001/03/10 06:03:36  spj
** Reduced amount of "hand coded" test setup in c-unit-suite because
** it was becoming onerous to repeat the code. W/e you find yourself
** copying and pasting code, think about perhaps making your life easier.
** --spj
**
** Revision 1.5  2001/03/08 06:35:50  spj
** Tested netstring functions.  Serialize is tested, unserialize needs to be
** written.
**
** Revision 1.4  2001/03/02 04:51:07  spj
** c-unit-suite moved from test to c-unit dir.; c-unit texinfo documentation renamed / re-outlined for future language support
**
** Revision 1.3  2001/02/23 05:10:10  spj
** Started a Texinfo manual document, and added the beginning of setup/teardown
** functions which are now invoked around the test.
**
** Revision 1.2  2001/02/19 19:26:14  spj
** 0.04 checkin.  Some report reformats and new unit tests for au_remove_test
** and au_delete_test.  TODO: au_delete_testcase.
** --spj
**
** Revision 1.1.1.1  2001/02/18 20:42:40  spj
** First checkin.  'c-unit' works but self-testing suite needs to be extended
** to cover more of its functionality. "self-testing" recursively with your
** own self-testing framework gets confusing sometimes. ;) --spj
**
**
*/

#include "test.h"
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/wait.h>

#ifdef AUTOUNIT_C_UNIT_DEBUG
#define au_log(str) g_log(G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "%s", str);
#else 
#define au_log(str) 
#endif

autounit_test_t*
au_new_test(GString *testName,
	    autounit_test_fp_t test_fp) {
  autounit_test_t *new_t = g_new(autounit_test_t,1);
  new_t->test_name = g_string_new(testName->str);
  new_t->test_status = g_string_new("");
  new_t->test_run = FALSE;
  new_t->failed_assertions = 0;
  new_t->test_seconds_elapsed = 0;
  new_t->test_useconds_elapsed = 0;
  new_t->test_fp = (gpointer) test_fp;
  new_t->do_fork = TRUE;
  return new_t;
}

GString *
au_test_serialize(autounit_test_t *t) {
  /* Convert a test structure into a stream of netstrings to pass back
  ** via the pipe to the test case parent process.  We omit pointers.
  ** This collection of netstrings itself then becomes one big
  ** netstring.  In the future, it might be the case that Autounit
  ** controls a farm of crosscompile machines... you never know. This
  ** is the simplest thing I can think of that will satisfy being
  ** reliable now and in the future.
  **/
  GString *str;
  GString *netstr;
  GString *tmp_test_name_ns, *tmp_test_run_ns, *tmp_test_status_ns, *tmp_failed_assertions_ns, 
    *tmp_test_seconds_elapsed_ns, *tmp_test_useconds_elapsed_ns;
  /* the fields we care about...
  GString *test_name;
  gboolean test_run;
  GString *test_status;
  gint     failed_assertions;
  gdouble  test_seconds_elapsed;
  gulong   test_useconds_elapsed;
  */
  
  tmp_test_name_ns = au_gstring_to_netstring(t->test_name);
  tmp_test_run_ns  = au_gboolean_to_netstring(t->test_run);
  tmp_test_status_ns = au_gstring_to_netstring(t->test_status);
  tmp_failed_assertions_ns = au_gint_to_netstring(t->failed_assertions);
  tmp_test_seconds_elapsed_ns = au_gdouble_to_netstring(t->test_seconds_elapsed);
  tmp_test_useconds_elapsed_ns = au_gulong_to_netstring(t->test_useconds_elapsed);

  str = g_string_new("");
  str = g_string_append(str,tmp_test_name_ns->str);
  str = g_string_append(str,tmp_test_run_ns->str);
  str = g_string_append(str,tmp_test_status_ns->str);
  str = g_string_append(str,tmp_failed_assertions_ns->str);
  str = g_string_append(str,tmp_test_seconds_elapsed_ns->str);
  str = g_string_append(str,tmp_test_useconds_elapsed_ns->str);
  netstr = au_gstring_to_netstring(str);

  g_string_free(str,TRUE);
  g_string_free(tmp_test_name_ns,TRUE);
  g_string_free(tmp_test_run_ns,TRUE);
  g_string_free(tmp_test_status_ns,TRUE);
  g_string_free(tmp_failed_assertions_ns,TRUE);
  g_string_free(tmp_test_seconds_elapsed_ns,TRUE);
  g_string_free(tmp_test_useconds_elapsed_ns,TRUE);

  return netstr;
}

autounit_test_t*
au_test_unserialize(GString *nst) {
  /* string,    boolean,  string,      int,               double,               ulong */
  /* test_name, test_run, test_status, failed_assertions, test_seconds_elapsed, test_useconds_elapsed */
  GString *test_name_ns, *test_run_ns, *test_status_ns, *failed_assertions_ns, *test_seconds_elapsed_ns,
    *test_useconds_elapsed_ns;
  GString *test_name, *test_status;

  autounit_test_t *us_test;
  GString *tmp_ns = g_string_new(nst->str);

  test_name_ns = au_pop_netstring(tmp_ns);
  test_run_ns = au_pop_netstring(tmp_ns);
  test_status_ns = au_pop_netstring(tmp_ns);
  failed_assertions_ns = au_pop_netstring(tmp_ns);
  test_seconds_elapsed_ns = au_pop_netstring(tmp_ns);
  test_useconds_elapsed_ns = au_pop_netstring(tmp_ns);

  test_name = au_netstring_to_gstring(test_name_ns);
  test_status = au_netstring_to_gstring(test_status_ns);
  us_test = au_new_test(test_name,0);
  us_test->test_run = au_netstring_to_gboolean(test_run_ns);
  g_string_append(us_test->test_status,test_status->str);
  us_test->failed_assertions = au_netstring_to_gint(failed_assertions_ns);
  us_test->test_seconds_elapsed = au_netstring_to_gdouble(test_seconds_elapsed_ns);
  us_test->test_useconds_elapsed = au_netstring_to_gulong(test_useconds_elapsed_ns);
  g_string_free(tmp_ns,TRUE);
  g_string_free(test_name,TRUE);
  g_string_free(test_status,TRUE);
  g_string_free(failed_assertions_ns,TRUE);
  g_string_free(test_seconds_elapsed_ns,TRUE);
  g_string_free(test_useconds_elapsed_ns,TRUE);
  return us_test;
}

void
au_set_fork_mode(autounit_test_t *t, gboolean mode)
{
  t->do_fork = mode;
}

void
au_run_test(autounit_test_t *t) {
  /* Because there was basically no difference between au_run_test and
  ** au_run_stress_test, I decided to make the debugged forking-au_run_test
  ** a private function and the other two some wrappers. */
  autounit_stress_report_t stress_report;
  gint i = 1;

  au_log("au_run_test invoked");
  stress_report.round = &i;
  stress_report.modulo = 1;

  au_run_test_fork(t,&stress_report);
}

void
au_run_stress_test(autounit_test_t *t,
		   autounit_stress_report_t *stress_report) {
  au_log("au_run_stress_test invoked");
  au_run_test_fork(t,stress_report);
}

static void
au_run_test_child(autounit_test_t *t)
{
  GTimer *elapsed_timer;

  elapsed_timer = g_timer_new();
  {
    autounit_test_fp_t su_fn = (autounit_test_setup_fp_t) t->testcase->setup_fp;
    autounit_test_fp_t fn = (autounit_test_fp_t) t->test_fp;
    autounit_test_fp_t td_fn = (autounit_test_teardown_fp_t) t->testcase->teardown_fp;
    if (su_fn) su_fn(t);
    fn(t);
    if (td_fn) td_fn(t);
  }
  g_timer_stop(elapsed_timer);
  t->test_seconds_elapsed = g_timer_elapsed(elapsed_timer,
					    &t->test_useconds_elapsed);
  t->test_run = TRUE;
  g_timer_destroy(elapsed_timer); 
}

void
au_run_test_fork(autounit_test_t *t,
		 autounit_stress_report_t *status) {
  /* start timer
  ** fork process
  ** in parent process:
  **   wait on child to return results and exit
  ** in new process:
  **   invoke setup_fp
  **   call test_fp
  **   call teardown_fp
  **   return results to parent process
  */
  int fork_result;
  int C2P_pipe[2];
  GString *parent_test_result;
  GString *child_test_result;
  autounit_test_t *tmp_t;

  au_log("au_run_test invoked");

  if (t->do_fork) {
    if (pipe(C2P_pipe) < 0) {
      au_assert(t,"pipe failed",0);
      return;
    }

    fork_result = fork();
    if (fork_result < 0) {
      au_assert(t,"fork failed",0);
      return;
    }
  
    if (fork_result == 0) {
      au_run_test_child(t);
      au_log("au_run_test returning");
      child_test_result = au_test_serialize(t);
      close(C2P_pipe[0]);
      write(C2P_pipe[1],child_test_result->str,child_test_result->len);
      close(C2P_pipe[1]);
      exit(0);
    }

    if (fork_result) { 
      /* Parent */
      gint status;
      close(C2P_pipe[1]);
      waitpid(fork_result,&status,0);
    
      if (WIFEXITED(status)) {
	parent_test_result = au_read_netstring(C2P_pipe[0]);
	tmp_t = au_test_unserialize(parent_test_result);
    
	/* Copy child's assertions to the parent's copy, we'll assume
	   that the test_name didn't change. */
	t->failed_assertions += tmp_t->failed_assertions;
	t->test_seconds_elapsed = tmp_t->test_seconds_elapsed;
	t->test_useconds_elapsed = tmp_t->test_useconds_elapsed;
	t->test_status = g_string_append(t->test_status,tmp_t->test_status->str);
      } else {
	if (WIFSIGNALED(status)) {
	  au_assert(t,"test failed with signal",0);	
	}
	if (WCOREDUMP(status)) {
	  au_assert(t,"test failed with core dump",0);	
	}
      }
    }
  }
  else
    au_run_test_child(t);
  
  if (t->failed_assertions > 0) {
    g_print("%c",'!');
  } else {
    if ((*(status->round) % status->modulo) == 0) {
      g_print("%c",'.');
    }
  }
}

void
au_remove_test(autounit_testcase_t *tc, autounit_test_t *t) {
  /* Remove first instance of test from testcase if it exists. */
  if (tc->tests != NULL) {
    tc->tests = g_slist_remove(tc->tests,t);
  }
}

void
au_delete_test(autounit_test_t *t) {
  /* Free's memory used by test, please don't call this on anything still linked into a 
     testcase, I'm trying to decide if it should call remove for the user before it deletes. */
  g_string_free(t->test_name,TRUE);
  g_string_free(t->test_status,TRUE);
  g_free(t);
}

autounit_testcase_t*
au_new_testcase(GString *testCaseName,
		autounit_test_setup_fp_t setup_fp,
		autounit_test_teardown_fp_t teardown_fp) {
  autounit_testcase_t *new_tc;
  au_log("au_new_testcase invoked");
  new_tc = g_new(autounit_testcase_t,1);
  new_tc->testcase_name = g_string_new(testCaseName->str);
  new_tc->testcase_status = g_string_new("");
  new_tc->testcase_run = FALSE;
  new_tc->testcase_pct_complete = 0;
  new_tc->testcase_seconds_elapsed = 0;
  new_tc->testcase_useconds_elapsed = 0;
  new_tc->setup_fp = (gpointer) setup_fp;
  new_tc->teardown_fp = (gpointer) teardown_fp;
  new_tc->tests = NULL;
  au_log("au_new_testcase returning");
  return new_tc;
}

autounit_testcase_t*
au_add_test(autounit_testcase_t* tc, autounit_test_t *t) {
  au_log("au_add_testcase invoked");
  t->testcase = tc;
  tc->tests = g_slist_append(tc->tests,t);
  au_log("au_add_testcase returning");
  return tc;
}

gboolean 
au_test_failed_check(autounit_testcase_t *tc) {
  gboolean failed_assertions_found;
  GSList* tmp;
  autounit_test_t *test;

  au_log("au_test_failed_check invoked");
  failed_assertions_found = FALSE;
  tmp = tc->tests;

  while (tmp != NULL)
    {
      test = (autounit_test_t*) tmp->data;
      if (test->failed_assertions > 0) {
	failed_assertions_found = TRUE;
      }
     tmp = g_slist_next(tmp);
   }  
  au_log("au_test_failed_check returning");
  return failed_assertions_found;
}

void
au_testcase_elapsed_time(autounit_testcase_t *tc) {
  /* Count up elapsed time from all tests in testcase and update
  ** summary.
  */
  GSList* tmp;
  autounit_test_t *test;

  au_log("au_testcase_elapsed_time invoked");
  tmp = tc->tests;
  while (tmp != NULL)
    {
      test = (autounit_test_t*) tmp->data;
      tc->testcase_seconds_elapsed += test->test_seconds_elapsed;
      tc->testcase_useconds_elapsed += test->test_useconds_elapsed;
      tmp = g_slist_next(tmp);
   }  
  au_log("au_testcase_elapsed_time returning");
}

void
au_testcase_reset_elapsed_time(autounit_testcase_t *tc) {
  /* Zero elapsed time for all tests in testcase and zero
  ** summary elapsed time.
  */
  GSList* tmp;
  autounit_test_t *test;

  au_log("au_testcase_reset_elapsed_time invoked");
  tmp = tc->tests;
  while (tmp != NULL)
    {
      test = (autounit_test_t*) tmp->data;
      test->test_seconds_elapsed = 0; 
      test->test_useconds_elapsed = 0;
      tmp = g_slist_next(tmp);
   }  
  tc->testcase_seconds_elapsed = 0;
  tc->testcase_useconds_elapsed = 0;
  au_log("au_testcase_reset_elapsed_time returning");
}

void
au_testcase_fail_report(autounit_testcase_t *tc) {
  GSList* tmp;
  gint test_num;
  autounit_test_t* test;

  au_log("au_testcase_fail_report invoked");
  tmp = tc->tests;
  test_num = 0;
  g_print("\nFAIL ");
  g_print("%s",tc->testcase_name->str);
  g_print("%f s (%d us) total elapsed time\n",
	  tc->testcase_seconds_elapsed,
	  tc->testcase_useconds_elapsed);
  while (tmp != NULL)
   {
     test = (autounit_test_t*)tmp->data;
     if (strlen(test->test_status->str)) {
       g_print("%s:",test->test_name->str);
       g_print("elapsed time %f s (%d us)",
	       test->test_seconds_elapsed,
	       test->test_useconds_elapsed);
       g_print(":failed assertions:\n%s",test->test_status->str);
     } 
     tmp = g_slist_next(tmp);
   }
  au_log("au_testcase_fail_report returning");
}

void
au_testcase_ok_report(autounit_testcase_t *tc) {
  au_log("au_testcase_ok_report invoked");
  g_print("\nOK ");
  g_print("%f s (%d us) total elapsed time\n",
	  tc->testcase_seconds_elapsed,
	  tc->testcase_useconds_elapsed);
  au_log("au_testcase_fail_report returning");
}

gint
au_run_testcase(autounit_testcase_t *tc) {
  gint result = 0;
  au_log("au_run_testcase invoked");
  au_testcase_reset_elapsed_time(tc);
  g_print("%s\n",tc->testcase_name->str);
  g_slist_foreach(tc->tests,(GFunc)au_run_test,0);
  g_print("\n");

  au_testcase_elapsed_time(tc);
  if (au_test_failed_check(tc) == TRUE) {
    au_testcase_fail_report(tc);
    result = 1;
  } else {
    au_testcase_ok_report(tc);
  }

  au_log("au_run_testcase returning");
  return result;
}

gint
au_run_stress_testcase(autounit_testcase_t *tc,
		       gint iterations,
		       gint status_modulo) {
  gint result = 0;
  gint i;
  autounit_stress_report_t stress_report;

  au_log("au_run_stress_testcase invoked");
  stress_report.round = &i;
  stress_report.modulo = status_modulo;

  g_print("%s (%d iterations)\n",tc->testcase_name->str,iterations);
  au_testcase_reset_elapsed_time(tc);
  for (i= 0; i < iterations; i++) {
    if ((*stress_report.round % stress_report.modulo) == 0)
      g_print("(%d)",i+1);

    g_slist_foreach(tc->tests,(GFunc)au_run_stress_test,&stress_report);
    au_testcase_elapsed_time(tc);
    if (au_test_failed_check(tc) == TRUE) {
      au_testcase_fail_report(tc);
      result = 1;
      /* Abort stress testing on first error */
      return result;
    }
    if (result) {
      break;
    }
  }
  g_print("\n");
  if (!result) {
    au_testcase_ok_report(tc);
  }else { 
    au_testcase_fail_report(tc);
  }
  au_log("au_run_stress_testcase returning");
  return result;
}

void
au_delete_testcase(autounit_testcase_t *tc) {
  /* Free's memory used by test, please don't call this on anything
     still linked into a testcase, I'm trying to decide if it should
     call remove for the user before it deletes. */
  g_string_free(tc->testcase_name,TRUE);
  g_string_free(tc->testcase_status,TRUE);
  g_free(tc);
}
